import contextlib
from dataclasses import asdict
from smtplib import SMTPException
from typing import Callable, cast
from uuid import UUID

import strawberry
import strawberry_django
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import PasswordChangeForm, SetPasswordForm
from django.core.exceptions import ObjectDoesNotExist
from django.core.signing import BadSignature, SignatureExpired
from django.db import transaction
from strawberry.types import Info
from strawberry.types.field import StrawberryField

from gqlauth.core.constants import Messages, TokenAction
from gqlauth.core.exceptions import (
    PasswordAlreadySetError,
    TokenScopeError,
    UserAlreadyVerified,
    UserNotVerified,
)
from gqlauth.core.types_ import GQLAuthError, GQLAuthErrors, MutationNormalOutput
from gqlauth.core.utils import (
    cast_to_status_user,
    get_payload_from_token,
    get_user,
    get_user_by_email,
    inject_fields,
    revoke_user_refresh_token,
)
from gqlauth.jwt.types_ import (
    ObtainJSONWebTokenInput,
    ObtainJSONWebTokenType,
    RefreshTokenType,
    RevokeRefreshTokenType,
    TokenType,
    VerifyTokenInput,
    VerifyTokenType,
)
from gqlauth.models import RefreshToken, UserStatus
from gqlauth.settings import gqlauth_settings as app_settings
from gqlauth.user.forms import EmailForm, PasswordLessRegisterForm, RegisterForm, UpdateAccountForm
from gqlauth.user.helpers import check_captcha, confirm_password
from gqlauth.user.signals import user_registered, user_verified

UserModel = get_user_model()


class BaseMixin:
    field: StrawberryField
    REQUIRE_VERIFICATION: bool = False
    resolve_mutation: Callable

    @classmethod
    def verification_check(cls, info: Info) -> None:
        if cls.REQUIRE_VERIFICATION:
            user = get_user(info)
            try:
                if not user.status.verified:  # type: ignore
                    raise GQLAuthError(code=GQLAuthErrors.NOT_VERIFIED)
            except AttributeError:  # anonymous user has no status.
                if user.is_anonymous:
                    raise GQLAuthError(code=GQLAuthErrors.UNAUTHENTICATED)


class __CaptchaNotifyImportError:
    @property
    def field(self):
        raise ImportError(
            "you probably don't have pillow installed you'll need to add the [captcha] extra."
        )


Captcha = __CaptchaNotifyImportError()

with contextlib.suppress(ImportError):
    from gqlauth.captcha.models import Captcha as CaptchaModel
    from gqlauth.captcha.types_ import CaptchaType

    class Captcha:  # type: ignore[no-redef]
        """Creates a brand-new captcha. Returns a base64 encoded string of the
        captcha. And uuid representing the captcha id in the database. When you
        will try to log in or register You will need submit that uuid With the
        user input.

        **The captcha will be invoked when the timeout expires**.
        """

        @strawberry_django.field(description=__doc__)
        def field(self) -> CaptchaType:
            return CaptchaModel.create_captcha()


class RegisterMixin(BaseMixin):
    """Register user with fields defined in the settings. If the email field of
    the user model is part of the registration fields (default), check if there
    is no user with that email.

    If it exists, it does not register the user, even if the email field
    is not defined as unique (default of the default django user model).

    When creating the user, it also creates a `UserStatus` related to
    that user, making it possible to track if the user is archived /
    verified.

    Send account verification email.

    If allowed to not verified users login, return token.
    """

    @strawberry.input
    @inject_fields(app_settings.REGISTER_MUTATION_FIELDS)
    class RegisterInput:
        if not app_settings.ALLOW_PASSWORDLESS_REGISTRATION:
            password1: str
            password2: str

        if app_settings.REGISTER_REQUIRE_CAPTCHA:
            identifier: UUID
            userEntry: str

    form = (
        PasswordLessRegisterForm if app_settings.ALLOW_PASSWORDLESS_REGISTRATION else RegisterForm
    )

    @classmethod
    def resolve_mutation(cls, info, input_: RegisterInput) -> MutationNormalOutput:
        if app_settings.LOGIN_REQUIRE_CAPTCHA:
            check_res = check_captcha(input_)
            if check_res != Messages.CAPTCHA_VALID:
                return MutationNormalOutput(success=False, errors={"captcha": check_res})
        email = getattr(input_, "email", False)
        try:
            with transaction.atomic():
                f = cls.form(asdict(input_))  # type: ignore
                if not f.is_valid():
                    return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
                user = f.save()
                if email:
                    send_activation = app_settings.SEND_ACTIVATION_EMAIL is True
                    send_password_set = (
                        app_settings.ALLOW_PASSWORDLESS_REGISTRATION is True
                        and app_settings.SEND_PASSWORD_SET_EMAIL is True
                    )
                    if send_activation:
                        user.status.send_activation_email(info)

                    if send_password_set:
                        user.status.send_password_set_email(info)

                user_registered.send(sender=cls, user=user)
                return MutationNormalOutput(success=True)
        except SMTPException:
            return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)


class VerifyAccountMixin(BaseMixin):
    """Verify user account.

    Receive the token that was sent by email. If the token is valid,
    make the user verified by making the `user.status.verified` field
    true.
    """

    @strawberry.input
    class VerifyAccountInput:
        token: str

    @classmethod
    def resolve_mutation(cls, info: Info, input_: VerifyAccountInput) -> MutationNormalOutput:
        try:
            UserStatus.verify(input_.token)
            return MutationNormalOutput(success=True)
        except UserAlreadyVerified:
            return MutationNormalOutput(success=False, errors=Messages.ALREADY_VERIFIED)
        except SignatureExpired:
            return MutationNormalOutput(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return MutationNormalOutput(success=False, errors=Messages.INVALID_TOKEN)


class ResendActivationEmailMixin(BaseMixin):
    """Sends activation email.

    It is called resend because theoretically the first activation email
    was sent when the user registered.

    If there is no user with the requested email, a successful response
    is returned.
    """

    @strawberry.input
    class ResendActivationEmailInput:
        email: str

    @classmethod
    def resolve_mutation(cls, info, input_: ResendActivationEmailInput) -> MutationNormalOutput:
        try:
            email = input_.email
            f = EmailForm({"email": email})
            if f.is_valid():
                user = get_user_by_email(email)
                user.status.resend_activation_email(info)
                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except ObjectDoesNotExist:
            return MutationNormalOutput(success=True)  # even if user is not registered
        except SMTPException:
            return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)
        except UserAlreadyVerified:
            return MutationNormalOutput(success=False, errors={"email": Messages.ALREADY_VERIFIED})


class SendPasswordResetEmailMixin(BaseMixin):
    """Send password reset email.

    For non verified users, send an activation email instead.

    If there is no user with the requested email, a successful response
    is returned.
    """

    @strawberry.input
    class SendPasswordResetEmailInput:
        email: str

    @classmethod
    def resolve_mutation(cls, info, input_: SendPasswordResetEmailInput) -> MutationNormalOutput:
        try:
            email = input_.email
            f = EmailForm({"email": email})
            if f.is_valid():
                user = get_user_by_email(email)
                user.status.send_password_reset_email(info, [email])
                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except ObjectDoesNotExist:
            return MutationNormalOutput(success=True)  # even if user is not registered
        except SMTPException:
            return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)
        except UserNotVerified:
            user = get_user_by_email(input_.email)
            try:
                user.status.resend_activation_email(info)
                return MutationNormalOutput(
                    success=False,
                    errors={"email": Messages.NOT_VERIFIED_PASSWORD_RESET},
                )
            except SMTPException:
                return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)


class PasswordResetMixin(BaseMixin):
    """Change user password without old password.

    Receive the token that was sent by email.

    If token and new passwords are valid, update user password and in
    case of using refresh tokens, revoke all of them.

    Also, if user has not been verified yet, verify it.
    """

    @strawberry.input
    class PasswordResetInput:
        token: str
        new_password1: str
        new_password2: str

    form = SetPasswordForm

    @classmethod
    def resolve_mutation(cls, _, input_: PasswordResetInput) -> MutationNormalOutput:
        try:
            payload = get_payload_from_token(
                input_.token,
                TokenAction.PASSWORD_RESET,
                app_settings.EXPIRATION_PASSWORD_RESET_TOKEN,
            )
            user = UserModel._default_manager.get(**payload)
            status: "UserStatus" = getattr(user, "status")  # noqa: B009
            f = cls.form(user, asdict(input_))  # type: ignore
            if f.is_valid():
                revoke_user_refresh_token(user)
                user = f.save()  # type: ignore
                if status.verified is False:
                    status.verified = True
                    status.save(update_fields=["verified"])
                    user_verified.send(sender=cls, user=user)

                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except SignatureExpired:
            return MutationNormalOutput(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return MutationNormalOutput(success=False, errors=Messages.INVALID_TOKEN)


class PasswordSetMixin(BaseMixin):
    """
    Set user password - for password-less registration

    Receive the token that was sent by email.

    If token and new passwords are valid, set
    user password and in case of using refresh
    tokens, revoke all of them.

    Also, if user has not been verified yet, verify it.
    """

    @strawberry.input
    class PasswordSetInput:
        token: str
        new_password1: str
        new_password2: str

    form = SetPasswordForm

    @classmethod
    def resolve_mutation(cls, _, input_: PasswordSetInput) -> MutationNormalOutput:
        try:
            token = input_.token
            payload = get_payload_from_token(
                token,
                TokenAction.PASSWORD_SET,
                app_settings.EXPIRATION_PASSWORD_SET_TOKEN,
            )
            user = UserModel._default_manager.get(**payload)
            f = cls.form(user, asdict(input_))  # type: ignore
            if f.is_valid():
                # Check if user has already set a password
                if user.has_usable_password():
                    raise PasswordAlreadySetError
                revoke_user_refresh_token(user)
                user = f.save()  # type: ignore
                status: "UserStatus" = getattr(user, "status")  # noqa: B009

                if status.verified is False:
                    status.verified = True
                    status.save(update_fields=["verified"])

                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except SignatureExpired:
            return MutationNormalOutput(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return MutationNormalOutput(success=False, errors=Messages.INVALID_TOKEN)
        except PasswordAlreadySetError:
            return MutationNormalOutput(success=False, errors=Messages.PASSWORD_ALREADY_SET)


class ObtainJSONWebTokenMixin(BaseMixin):
    """Obtain JSON web token for given user.

    Allow to perform login with different fields, The fields are defined
    on settings.

    Not verified users can log in by default. This can be changes on
    settings.

    If user is archived, make it unarchived and return
    `unarchiving=True` on OutputBase.
    """

    @classmethod
    def resolve_mutation(cls, info, input_: ObtainJSONWebTokenInput) -> ObtainJSONWebTokenType:
        if app_settings.LOGIN_REQUIRE_CAPTCHA and not app_settings.CI_MODE:
            check_res = check_captcha(input_)
            if check_res != Messages.CAPTCHA_VALID:
                return ObtainJSONWebTokenType(success=False, errors={"captcha": check_res})

        return ObtainJSONWebTokenType.authenticate(info, input_)


class ArchiveOrDeleteMixin(BaseMixin):
    @strawberry.input
    class ArchiveOrDeleteMixinInput:
        password: str

    REQUIRE_VERIFICATION = True

    @classmethod
    def resolve_mutation(cls, info, input_: ArchiveOrDeleteMixinInput) -> MutationNormalOutput:
        user = get_user(info)
        if error := confirm_password(user, input_):
            return error
        cls.resolve_action(user)  # type: ignore
        return MutationNormalOutput(success=True)


class ArchiveAccountMixin(ArchiveOrDeleteMixin):
    """Archive account and revoke refresh tokens.

    User must be verified and confirm password.
    """

    @classmethod
    def resolve_action(cls, user):
        UserStatus.archive(user)
        revoke_user_refresh_token(user=user)


class DeleteAccountMixin(ArchiveOrDeleteMixin):
    """Delete account permanently or make `user.is_active=False`.

    The behavior is defined on settings. Anyway user refresh tokens are
    revoked.

    User must be verified and confirm password.
    """

    @classmethod
    def resolve_action(cls, user):
        if app_settings.ALLOW_DELETE_ACCOUNT:
            revoke_user_refresh_token(user=user)
            user.delete()


class PasswordChangeMixin(BaseMixin):
    """Change account password when user knows the old password.

    A new token and refresh token are sent. User must be verified.
    """

    @strawberry.input
    class PasswordChangeInput:
        old_password: str
        new_password1: str
        new_password2: str

    form = PasswordChangeForm
    REQUIRE_VERIFICATION = True

    @classmethod
    def resolve_mutation(cls, info: Info, input_: PasswordChangeInput) -> ObtainJSONWebTokenType:
        user = get_user(info)
        if error := confirm_password(user, input_):
            return ObtainJSONWebTokenType(**asdict(error))  # type: ignore

        args = asdict(input_)  # type: ignore
        f = cls.form(user, args)  # type: ignore
        if f.is_valid():
            revoke_user_refresh_token(user)
            user = f.save()
            user_with_status = cast_to_status_user(user)
            return ObtainJSONWebTokenType.from_user(user_with_status)
        else:
            return ObtainJSONWebTokenType(success=False, errors=f.errors.get_json_data())


class UpdateAccountMixin(BaseMixin):
    """Update user model fields, defined on settings.

    User must be verified.
    """

    @strawberry.input
    @inject_fields(app_settings.UPDATE_MUTATION_FIELDS)
    class UpdateAccountInput: ...

    form = UpdateAccountForm
    REQUIRE_VERIFICATION = True

    @classmethod
    def resolve_mutation(cls, info, input_: UpdateAccountInput) -> MutationNormalOutput:
        user = get_user(info)
        f = cls.form(
            asdict(input_),  # type: ignore
            instance=user,
        )
        if f.is_valid():
            f.save()
            return MutationNormalOutput(success=True)
        else:
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())


class VerifyTokenMixin(BaseMixin):
    """### Checks if a token is not expired and correct.

    *Note that this is not for refresh tokens.*
    """

    @classmethod
    def resolve_mutation(cls, _: Info, input_: VerifyTokenInput) -> VerifyTokenType:
        return VerifyTokenType.from_token(input_)


class RefreshTokenMixin(BaseMixin):
    """### refreshToken to generate a new login token:

    *Use this only if `JWT_LONG_RUNNING_REFRESH_TOKEN` is True*

    using the refresh-token you already got during authorization, and
    obtain a brand-new token (and possibly a new refresh token if you
    revoked the previous). This is an alternative to log in when your
    token expired.
    """

    @strawberry.input(
        description="If `revoke_refresh_token` is true,"
        " revokes to refresh token an creates a new one."
    )
    class RefreshTokenInput:
        refresh_token: str
        revoke_refresh_token: bool = strawberry.field(
            description="revokes the previous refresh token, and will generate new one.",
            default=False,
        )

    @classmethod
    def resolve_mutation(cls, info, input_: RefreshTokenInput) -> ObtainJSONWebTokenType:
        try:
            res = RefreshToken.objects.get(token=input_.refresh_token)
        except RefreshToken.DoesNotExist:
            return ObtainJSONWebTokenType(success=False, errors=Messages.INVALID_TOKEN)
        user = res.user
        if res.is_expired_():
            return ObtainJSONWebTokenType(success=False, errors=Messages.EXPIRED_TOKEN)
        # fields that are determined by if statements are not recognized by mypy.
        ret = ObtainJSONWebTokenType(
            success=True, token=TokenType.from_user(user), refresh_token=res  # type: ignore
        )
        if input_.revoke_refresh_token:
            res.revoke()
            ret.refresh_token = cast(RefreshTokenType, RefreshToken.from_user(user))
        return ret


class RevokeTokenMixin(BaseMixin):
    """### Suspends a refresh token.

    *token must exist to be revoked.*
    """

    @strawberry.input
    class RevokeTokenInput:
        refresh_token: str

    @classmethod
    def resolve_mutation(cls, _: Info, input_: RevokeTokenInput) -> RevokeRefreshTokenType:
        try:
            refresh_token = RefreshToken.objects.get(
                token=input_.refresh_token, revoked__isnull=True
            )
            refresh_token.revoke()
            return RevokeRefreshTokenType(
                success=True, refresh_token=cast(RefreshTokenType, refresh_token)
            )
        except RefreshToken.DoesNotExist:
            return RevokeRefreshTokenType(success=False, errors=Messages.INVALID_TOKEN)



from __future__ import annotations

import dataclasses
from typing import Any, Callable, cast, final

import strawberry
from graphql.pyutils import AwaitableOrValue
from strawberry.schema_directive import Location
from strawberry.types import Info
from strawberry_django.permissions import DjangoNoPermission, DjangoPermissionExtension
from strawberry_django.utils.typing import UserType
from typing_extensions import override

from gqlauth.core.utils import UserProto


@strawberry.schema_directive(
    locations=[Location.FIELD_DEFINITION],
    description="Checks whether a user is verified",
)
@final
class IsVerified(DjangoPermissionExtension):
    """Mark a field as only resolvable by authenticated users."""

    message: strawberry.Private[str] = dataclasses.field(default="User is not authenticated.")

    @override
    def resolve_for_user(
        self, resolver: Callable[..., Any], user: UserType | None, *, info: Info, source: Any
    ) -> AwaitableOrValue:
        user = cast(UserProto, user)
        if user.is_authenticated and user.status.verified:
            return resolver()
        raise DjangoNoPermission



import dataclasses
from datetime import datetime
from typing import TYPE_CHECKING, Optional, cast
from uuid import UUID

import strawberry
import strawberry_django
from django.contrib.auth import authenticate
from django.core.exceptions import PermissionDenied
from django.utils.timezone import localtime
from strawberry import auto
from strawberry.types import Info

from gqlauth.core.constants import Messages
from gqlauth.core.exceptions import TokenExpired
from gqlauth.core.interfaces import OutputInterface
from gqlauth.core.scalars import ExpectedErrorType
from gqlauth.core.utils import USER_MODEL, app_settings, inject_fields, utc_now
from gqlauth.models import RefreshToken
from gqlauth.user.types_ import UserType

if TYPE_CHECKING:
    from gqlauth.core.utils import UserProto


@strawberry_django.type(
    model=RefreshToken,
    description="""
Refresh token can be used to obtain a new token instead of log in again
when the token expires.

*This is only used if `JWT_LONG_RUNNING_REFRESH_TOKEN` is set to True.*
""",
)
class RefreshTokenType:
    token: auto = strawberry_django.field(
        description="randomly generated token that is attached to a FK user."
    )
    created: auto
    revoked: auto

    @strawberry_django.field
    def expires_at(self) -> datetime:
        self: RefreshToken  # type: ignore
        return self.expires_at_()  # type: ignore

    @strawberry_django.field
    def is_expired(self) -> bool:
        self: RefreshToken  # type: ignore
        return self.is_expired_()  # type: ignore


@strawberry.type(
    description="""
the data that was used to create the token.
"""
)
@inject_fields(
    {
        app_settings.JWT_PAYLOAD_PK,
    }
)
class TokenPayloadType:
    origIat: datetime = strawberry.field(
        description="when the token was created", default_factory=utc_now
    )
    exp: datetime = strawberry.field(description="when the token will be expired", default=None)

    def __post_init__(self):
        if not self.exp:
            self.exp = self.origIat + app_settings.JWT_EXPIRATION_DELTA

    def as_dict(self):
        ret = dataclasses.asdict(self)
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if isinstance(value, datetime):
                ret[field.name] = value.strftime(app_settings.JWT_TIME_FORMAT)
        return ret

    @classmethod
    def from_dict(cls, data: dict) -> "TokenPayloadType":
        for field in dataclasses.fields(cls):  # type: ignore
            value = data[field.name]
            if isinstance(value, str) and field.type is datetime:
                data[field.name] = datetime.strptime(value, app_settings.JWT_TIME_FORMAT)
        return cls(**data)


@strawberry.type(
    description="""
encapsulates the token with the payload that was used to create the token.
"""
)
class TokenType:
    payload: TokenPayloadType
    token: str = strawberry.field(description="The encoded payload, namely a token.")

    def is_expired(self):
        return self.payload.exp < utc_now()

    @classmethod
    def from_user(cls, user: "UserProto") -> "TokenType":
        return app_settings.JWT_PAYLOAD_HANDLER(user)

    @classmethod
    def from_token(cls, token: str) -> "TokenType":
        """Might raise TokenExpired."""
        token_type: TokenType = app_settings.JWT_DECODE_HANDLER(token)
        if token_type.is_expired():
            raise TokenExpired
        return token_type

    def get_user_instance(self) -> "UserProto":
        """Might raise not existed exception."""
        pk_name = app_settings.JWT_PAYLOAD_PK.python_name
        query = {pk_name: getattr(self.payload, pk_name)}
        return USER_MODEL.objects.get(**query)  # type: ignore


@strawberry.input
@inject_fields(app_settings.LOGIN_FIELDS)
class ObtainJSONWebTokenInput:
    password: str
    if app_settings.LOGIN_REQUIRE_CAPTCHA:
        identifier: UUID
        userEntry: str


@strawberry.type(
    description="""
    encapsulates token data, and refresh token data if `JWT_LONG_RUNNING_REFRESH_TOKEN` is on.
    with an output interface.
    """
)
class ObtainJSONWebTokenType(OutputInterface):
    success: bool
    user: Optional[UserType] = None
    token: Optional[TokenType] = None
    if app_settings.JWT_LONG_RUNNING_REFRESH_TOKEN:
        refresh_token: Optional[RefreshTokenType] = None
    errors: Optional[ExpectedErrorType] = None

    @classmethod
    def from_user(cls, user: "UserProto") -> "ObtainJSONWebTokenType":
        """Creates a new token and possibly a new refresh token based on the
        user.

        *call this method only for trusted users.*
        """
        ret = ObtainJSONWebTokenType(
            success=True, user=cast(UserType, user), token=TokenType.from_user(user)
        )
        if app_settings.JWT_LONG_RUNNING_REFRESH_TOKEN:
            ret.refresh_token = cast(RefreshTokenType, RefreshToken.from_user(user))
        return ret

    @classmethod
    def authenticate(cls, info: Info, input_: ObtainJSONWebTokenInput) -> "ObtainJSONWebTokenType":
        """Return `ObtainJSONWebTokenType`. authenticates against django
        authentication backends.

        *creates a new token and possibly a refresh token.*
        """
        args = {
            USER_MODEL.USERNAME_FIELD: getattr(input_, USER_MODEL.USERNAME_FIELD),
            "password": input_.password,
        }
        try:
            # authenticate against django authentication backends.
            user: Optional["UserProto"]
            request = (
                info.context["request"] if isinstance(info.context, dict) else info.context.request
            )
            if not (user := authenticate(request, **args)):  # type: ignore
                return ObtainJSONWebTokenType(success=False, errors=Messages.INVALID_CREDENTIALS)

            from gqlauth.models import UserStatus

            status: UserStatus = getattr(user, "status")  # noqa: B009
            # gqlauth logic
            if status.archived is True:  # un-archive on login
                UserStatus.unarchive(user)
            if status.verified or app_settings.ALLOW_LOGIN_NOT_VERIFIED:
                # successful login.
                user.last_login = localtime()
                user.save(update_fields=("last_login",))
                return ObtainJSONWebTokenType.from_user(user)
            else:
                return ObtainJSONWebTokenType(success=False, errors=Messages.NOT_VERIFIED)

        except PermissionDenied:
            # one of the authentication backends rejected the user.
            return ObtainJSONWebTokenType(success=False, errors=Messages.UNAUTHENTICATED)

        except TokenExpired:
            return ObtainJSONWebTokenType(success=False, errors=Messages.EXPIRED_TOKEN)


@strawberry.input
class VerifyTokenInput:
    token: str


@strawberry.type
class VerifyTokenType(OutputInterface):
    success: bool
    token: Optional[TokenType] = None
    user: Optional[UserType] = None
    errors: Optional[ExpectedErrorType] = None

    @classmethod
    def from_token(cls, token_input: VerifyTokenInput) -> "VerifyTokenType":
        try:
            token_type = TokenType.from_token(token_input.token)
            user = token_type.get_user_instance()
        except USER_MODEL.DoesNotExist:
            return VerifyTokenType(success=False, errors=Messages.INVALID_CREDENTIALS)
        except TokenExpired:
            return VerifyTokenType(success=False, errors=Messages.EXPIRED_TOKEN)

        else:
            return VerifyTokenType(token=token_type, user=cast(UserType, user), success=True)


@strawberry.type
class RevokeRefreshTokenType:
    success: bool
    refresh_token: Optional[RefreshTokenType] = None
    errors: Optional[ExpectedErrorType] = None



import json
from dataclasses import dataclass, field
from datetime import timedelta
from random import SystemRandom
from typing import (
    TYPE_CHECKING,
    Callable,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from django.http.request import HttpRequest
from strawberry import ID
from strawberry.annotation import StrawberryAnnotation
from strawberry.types.field import StrawberryField

import jwt

if TYPE_CHECKING:  # pragma: no cover
    from django.contrib.auth.base_user import AbstractBaseUser

    from gqlauth.core.utils import UserProto
    from gqlauth.jwt.types_ import TokenType


def token_finder(request_or_scope: Union[dict, "HttpRequest"]) -> Optional[str]:
    from gqlauth.core.constants import JWT_PREFIX

    token: Optional[str] = None
    if isinstance(request_or_scope, HttpRequest):  # django
        headers = request_or_scope.headers
        token = headers.get("authorization", None) or headers.get("Authorization", None)
    else:  # channels
        request_or_scope = cast(dict, request_or_scope)
        raw_headers: List[Tuple[bytes, bytes]] = request_or_scope["headers"]
        for k, v in raw_headers:
            if k == b"authorization":
                token = v.decode()
                break
    if token and token.startswith(JWT_PREFIX):
        return token[len(JWT_PREFIX) :].strip(" ")
    return None


def create_token_type(user: "AbstractBaseUser") -> "TokenType":
    from gqlauth.core.utils import app_settings
    from gqlauth.jwt.types_ import TokenPayloadType, TokenType

    user_pk = app_settings.JWT_PAYLOAD_PK.python_name
    pk_field = {user_pk: getattr(user, user_pk)}
    payload = TokenPayloadType(
        **pk_field,
    )
    serialized = json.dumps(payload.as_dict(), sort_keys=True, indent=1)
    return TokenType(
        token=str(
            jwt.encode(
                payload={"payload": serialized},
                key=cast(str, app_settings.JWT_SECRET_KEY.value),
                algorithm=app_settings.JWT_ALGORITHM,
            )
        ),
        payload=payload,
    )


def decode_jwt(token: str) -> "TokenType":
    from gqlauth.core.utils import app_settings
    from gqlauth.jwt.types_ import TokenPayloadType, TokenType

    decoded = json.loads(
        jwt.decode(
            token,
            key=cast(str, app_settings.JWT_SECRET_KEY.value),
            algorithms=[
                app_settings.JWT_ALGORITHM,
            ],
        )["payload"]
    )
    return TokenType(token=token, payload=TokenPayloadType.from_dict(decoded))


def default_text_factory():
    return "".join(
        [
            str(node)
            for node in [
                SystemRandom().randint(0, 10) for _ in range(5, SystemRandom().randint(10, 20))
            ]
        ]
    )


T = TypeVar("T")


class DjangoSetting(Generic[T]):
    __slots__ = ("setting", "cached")

    def __init__(self, setting: str, value: Optional[T] = None):
        self.setting = setting
        self.cached: Optional[T] = value

    @property
    def value(self) -> T:
        if self.cached:  # slotted classes can't use cached property (without __dict__)
            return self.cached
        from django.conf import settings

        final = settings
        for attr in self.setting.split("."):
            final = getattr(final, attr)

        return final  # type: ignore

    @classmethod
    def override(cls, value: T) -> "DjangoSetting":
        return DjangoSetting(setting="", value=value)


id_field = StrawberryField(python_name="id", default=None, type_annotation=StrawberryAnnotation(ID))
username_field = StrawberryField(
    python_name="username", default=None, type_annotation=StrawberryAnnotation(str)
)
password_field = StrawberryField(
    python_name="password", default=None, type_annotation=StrawberryAnnotation(str)
)
first_name_field = StrawberryField(
    python_name="first_name",
    default=None,
    type_annotation=StrawberryAnnotation(Optional[str]),
)
last_name_field = StrawberryField(
    python_name="last_name",
    default=None,
    type_annotation=StrawberryAnnotation(Optional[str]),
)
email_field = StrawberryField(
    python_name="email", default=None, type_annotation=StrawberryAnnotation(str)
)


def default_captcha_text_validator(original: str, received: str) -> bool:
    return original == received


@dataclass
class GqlAuthSettings:
    ALLOW_LOGIN_NOT_VERIFIED: bool = False
    """"""
    LOGIN_FIELDS: Set[StrawberryField] = field(
        default_factory=lambda: {
            username_field,
        }
    )
    """These fields would be used to authenticate with SD-jwt `authenticate`
    function. This function will call each of our `AUTHENTICATION_BACKENDS`,
    And will return the user from one of them unless `PermissionDenied` was
    raised. You can pass any fields that would be accepted by your backends.

    **Note that `password field` is mandatory and cannot be removed.**
    """
    LOGIN_REQUIRE_CAPTCHA: bool = True
    """Whether login will require captcha verification."""
    CI_MODE: bool = False
    """Whether to enable CI mode, this will disable captcha validation
    (although the fields would still be required)."""
    REGISTER_MUTATION_FIELDS: Set[StrawberryField] = field(
        default_factory=lambda: {email_field, username_field}
    )
    """Fields on register, plus password1 and password2, can be a dict like
    UPDATE_MUTATION_fieldS setting."""
    REGISTER_REQUIRE_CAPTCHA: bool = True
    """Whether register will require captcha verification."""
    # captcha stuff
    #: captcha expiration delta.
    CAPTCHA_EXPIRATION_DELTA: timedelta = timedelta(seconds=120)
    #: max number of attempts for one captcha.
    CAPTCHA_MAX_RETRIES: int = 5
    CAPTCHA_TEXT_FACTORY: Callable = default_text_factory
    """A callable with no arguments that returns a string.

    This will be used to generate the captcha image.
    """
    CAPTCHA_TEXT_VALIDATOR: Callable[[str, str], bool] = default_captcha_text_validator
    """A callable that will receive the original string vs user input and
    returns a boolean."""
    FORCE_SHOW_CAPTCHA: bool = False
    """Whether to show the captcha image after it has been created for
    debugging purposes."""
    CAPTCHA_SAVE_IMAGE: bool = False
    """If True, an png representation of the captcha will be saved under
    MEDIA_ROOT/captcha/<datetime>/<uuid>.png."""
    # optional fields on update account, can be list of fields
    UPDATE_MUTATION_FIELDS: Set[StrawberryField] = field(
        default_factory=lambda: {first_name_field, last_name_field}
    )
    """Fields on update account mutation."""

    # email tokens
    EXPIRATION_ACTIVATION_TOKEN: timedelta = timedelta(days=7)
    EXPIRATION_PASSWORD_RESET_TOKEN: timedelta = timedelta(hours=1)
    EXPIRATION_PASSWORD_SET_TOKEN: timedelta = timedelta(hours=1)
    # email stuff

    EMAIL_FROM: DjangoSetting[str] = DjangoSetting("DEFAULT_FROM_EMAIL")
    SEND_ACTIVATION_EMAIL: bool = True
    ACTIVATION_PATH_ON_EMAIL: str = "activate"
    PASSWORD_SET_PATH_ON_EMAIL: str = "password-set"
    PASSWORD_RESET_PATH_ON_EMAIL: str = "password-reset"
    # email subjects templates
    EMAIL_SUBJECT_ACTIVATION: str = "email/activation_subject.txt"
    EMAIL_SUBJECT_ACTIVATION_RESEND: str = "email/activation_subject.txt"
    EMAIL_SUBJECT_PASSWORD_SET: str = "email/password_set_subject.txt"
    EMAIL_SUBJECT_PASSWORD_RESET: str = "email/password_reset_subject.txt"
    # email templates
    EMAIL_TEMPLATE_ACTIVATION: str = "email/activation_email.html"
    EMAIL_TEMPLATE_ACTIVATION_RESEND: str = "email/activation_email.html"
    EMAIL_TEMPLATE_PASSWORD_SET: str = "email/password_set_email.html"
    EMAIL_TEMPLATE_PASSWORD_RESET: str = "email/password_reset_email.html"
    EMAIL_TEMPLATE_VARIABLES: dict = field(default_factory=lambda: {})
    # others
    ALLOW_DELETE_ACCOUNT: bool = False
    """If True, DeleteAcount mutation will permanently delete the user."""
    ALLOW_PASSWORDLESS_REGISTRATION: bool = False
    """Whether to allow registration with no password."""
    SEND_PASSWORD_SET_EMAIL: bool = False
    # JWT stuff
    JWT_SECRET_KEY: DjangoSetting[str] = DjangoSetting("SECRET_KEY")
    """Key used to sign the JWT token."""

    JWT_ALGORITHM: str = "HS256"
    """Algorithm used for signing the token."""
    JWT_TIME_FORMAT: str = "%Y-%m-%d-%H-%M-%S-%z"
    """A valid 'strftime' string that will be used to encode the token
    payload."""
    JWT_PAYLOAD_HANDLER: Callable[["UserProto"], "TokenType"] = create_token_type
    """A custom function to generate the token datatype, its up to you to
    encode the token."""
    JWT_PAYLOAD_PK: StrawberryField = field(default_factory=lambda: username_field)
    """Field that will be used to generate the token from a user instance and
    retrieve user based on the decoded token. The default value is the username
    field of the User model. If you want to change it to id, for example, you
    can use the id_field defined in gqlauth.settings_type and change it like
    JWT_PAYLOAD_PK=id_field.

    *This filed must be unique in the database*
    """
    JWT_DECODE_HANDLER: Callable[[str], "TokenType"] = decode_jwt

    JWT_TOKEN_FINDER: Callable[[Union["HttpRequest", dict]], Optional[str]] = token_finder
    """A hook called by `GqlAuthRootField` to find the token. Accepts the
    request object (might be channels scope dict or django request object)

    **remember to strip the "JWT " prefix if you override this.**
    """
    JWT_EXPIRATION_DELTA: timedelta = timedelta(minutes=5)
    """Timedelta added to `utcnow()` to set the expiration time.

    When this ends you will have to create a new token by logging in or
    using the refresh token.
    """
    # refresh token stuff
    JWT_LONG_RUNNING_REFRESH_TOKEN: bool = True
    """Whether to enable refresh tokens to be used as an alternative to login
    every time the token is expired."""
    JWT_REFRESH_TOKEN_N_BYTES: int = 20
    """Number of bytes for long running refresh token."""
    JWT_REFRESH_EXPIRATION_DELTA: timedelta = timedelta(days=7)
    """Refresh token expiration time delta."""

    def __post_init__(self):
        # if there override the defaults
        if "email" not in {field_.name for field_ in self.REGISTER_MUTATION_FIELDS}:
            self.SEND_ACTIVATION_EMAIL = False



import binascii
import os
import time
from datetime import datetime

from django.conf import settings as django_settings
from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import send_mail
from django.db import models
from django.db.models import Case, When
from django.db.models import Value as Val
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.html import strip_tags
from django.utils.translation import gettext_lazy as _
from strawberry.types import Info

from gqlauth.core.constants import TokenAction
from gqlauth.core.exceptions import UserAlreadyVerified
from gqlauth.core.utils import get_payload_from_token, get_token

# gqlauth imports
from gqlauth.settings import gqlauth_settings as app_settings
from gqlauth.user.signals import user_verified

USER_MODEL = get_user_model()


class UserStatus(models.Model):
    """A helper model that handles user account stuff."""

    class Meta:
        verbose_name_plural = "User statuses"

    user = models.OneToOneField(
        django_settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="status"
    )
    verified = models.BooleanField(default=False)
    archived = models.BooleanField(default=False)

    def __str__(self):
        return "%s - status" % (self.user)

    def send(self, subject, template, context, recipient_list=None):
        _subject = render_to_string(subject, context).replace("\n", " ").strip()
        html_message = render_to_string(template, context)
        message = strip_tags(html_message)

        return send_mail(
            subject=_subject,
            from_email=app_settings.EMAIL_FROM.value,
            message=message,
            html_message=html_message,
            recipient_list=(recipient_list or [getattr(self.user, USER_MODEL.EMAIL_FIELD)]),
            fail_silently=False,
        )

    def get_email_context(self, info: Info, path, action, **kwargs):
        token = get_token(self.user, action, **kwargs)
        is_channels = isinstance(info.context, dict)
        if is_channels:
            request = info.context["request"]
            domain, port = request.headers["host"].split(":")
            site_name = domain
            protocol = request.consumer.scope["type"]
        else:
            request = info.context.request
            site = get_current_site(request)
            port = request.get_port()
            site_name = site.name
            domain = site.domain
            protocol = "https" if request.is_secure() else "http"

        return {
            "user": self.user,
            "request": request,
            "token": token,
            "port": port,
            "site_name": site_name,
            "domain": domain,
            "protocol": protocol,
            "path": path,
            "timestamp": time.time(),
            **app_settings.EMAIL_TEMPLATE_VARIABLES,
        }

    def send_activation_email(self, info, *args, **kwargs):
        email_context = self.get_email_context(
            info, app_settings.ACTIVATION_PATH_ON_EMAIL, TokenAction.ACTIVATION
        )
        template = app_settings.EMAIL_TEMPLATE_ACTIVATION
        subject = app_settings.EMAIL_SUBJECT_ACTIVATION
        return self.send(subject, template, email_context, *args, **kwargs)

    def resend_activation_email(self, info, *args, **kwargs):
        if self.verified is True:
            raise UserAlreadyVerified
        email_context = self.get_email_context(
            info, app_settings.ACTIVATION_PATH_ON_EMAIL, TokenAction.ACTIVATION
        )
        template = app_settings.EMAIL_TEMPLATE_ACTIVATION_RESEND
        subject = app_settings.EMAIL_SUBJECT_ACTIVATION_RESEND
        return self.send(subject, template, email_context, *args, **kwargs)

    def send_password_set_email(self, info, *args, **kwargs):
        email_context = self.get_email_context(
            info, app_settings.PASSWORD_SET_PATH_ON_EMAIL, TokenAction.PASSWORD_SET
        )
        template = app_settings.EMAIL_TEMPLATE_PASSWORD_SET
        subject = app_settings.EMAIL_SUBJECT_PASSWORD_SET
        return self.send(subject, template, email_context, *args, **kwargs)

    def send_password_reset_email(self, info, *args, **kwargs):
        email_context = self.get_email_context(
            info, app_settings.PASSWORD_RESET_PATH_ON_EMAIL, TokenAction.PASSWORD_RESET
        )
        template = app_settings.EMAIL_TEMPLATE_PASSWORD_RESET
        subject = app_settings.EMAIL_SUBJECT_PASSWORD_RESET
        return self.send(subject, template, email_context, *args, **kwargs)

    @classmethod
    def verify(cls, token):
        payload = get_payload_from_token(
            token, TokenAction.ACTIVATION, app_settings.EXPIRATION_ACTIVATION_TOKEN
        )
        user = USER_MODEL._default_manager.get(**payload)
        user_status = cls.objects.get(user=user)
        if user_status.verified is False:
            user_status.verified = True
            user_status.save(update_fields=["verified"])
            user_verified.send(sender=cls, user=user)
        else:
            raise UserAlreadyVerified

    @classmethod
    def unarchive(cls, user):
        user_status = cls.objects.get(user=user)
        if user_status.archived is True:
            user_status.archived = False
            user_status.save(update_fields=["archived"])

    @classmethod
    def archive(cls, user):
        user_status = cls.objects.get(user=user)
        if user_status.archived is False:
            user_status.archived = True
            user_status.save(update_fields=["archived"])


class RefreshTokenQuerySet(models.QuerySet):
    def expired(self):
        expires = timezone.now() - app_settings.JWT_REFRESH_EXPIRATION_DELTA
        return self.annotate(
            expired=Case(
                When(created__lt=expires, then=Val(True)),
                output_field=models.BooleanField(),
                default=Val(False),
            ),
        )


class RefreshToken(models.Model):
    """Refresh token is a random set of bytes decoded to a string that is
    referring a user.

    It can be used to retrieve a new token without the need to login
    again.
    """

    id = models.BigAutoField(primary_key=True)  # noqa A003
    user = models.ForeignKey(
        USER_MODEL,
        on_delete=models.CASCADE,
        related_name="refresh_tokens",
        verbose_name=_("user"),
    )
    token = models.CharField(_("token"), max_length=255, editable=False)
    created = models.DateTimeField(_("created"), auto_now_add=True)
    revoked = models.DateTimeField(_("revoked"), null=True, blank=True)

    def expires_at_(self) -> datetime:
        return self.created + app_settings.JWT_REFRESH_EXPIRATION_DELTA

    def is_expired_(self) -> bool:
        """Whether the token is expired or not.

        it is up to the database query to filter out tokens without
        revoked date.
        """
        return datetime.now(tz=self.created.tzinfo) > self.expires_at_() or bool(self.revoked)

    def revoke(self):
        self.revoked = timezone.now()
        self.save(update_fields=["revoked"])

    objects = RefreshTokenQuerySet.as_manager()  # type: ignore

    class Meta:
        verbose_name = _("refresh token")
        verbose_name_plural = _("refresh tokens")
        unique_together = ("token", "revoked")

    def __str__(self):
        return self.token

    @classmethod
    def from_user(cls, user) -> "RefreshToken":
        token = binascii.hexlify(
            os.urandom(app_settings.JWT_REFRESH_TOKEN_N_BYTES),
        ).decode()

        obj = RefreshToken.objects.create(user=user, token=token)
        obj.save()
        return obj



import contextlib
import datetime
import inspect
import typing
from typing import Dict, Iterable

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import AnonymousUser
from django.core import signing
from strawberry.types import Info
from strawberry.types.field import StrawberryField
from strawberry.utils.str_converters import to_camel_case

from gqlauth.core.exceptions import TokenScopeError

# Without this it would cause django to complain about import ordering.
if typing.TYPE_CHECKING:  # pragma: no cover
    from gqlauth.models import UserStatus
    from gqlauth.settings_type import GqlAuthSettings

    class UserProto(AbstractBaseUser):
        status: UserStatus

else:
    UserProto = "Foobar to allow import me in runtime."

USER_MODEL = get_user_model()
USER_UNION = typing.Union["UserProto", AnonymousUser, AbstractBaseUser]
app_settings: "GqlAuthSettings" = settings.GQL_AUTH


def hide_args_kwargs(field):
    sig = inspect.signature(field)
    cleared = tuple(p for p in sig.parameters.values() if p.name not in ("kwargs", "args"))
    field.__signature__ = inspect.signature(field).replace(parameters=(cleared))
    return field


def isiterable(value):
    try:
        iter(value)
    except TypeError:  # pragma: no cover
        return False
    return True


def camelize(data):
    if isinstance(data, dict):
        return {to_camel_case(k): camelize(v) for k, v in data.items()}
    if isiterable(data) and not isinstance(data, str):
        return [camelize(d) for d in data]
    return data


def get_user(info: Info) -> USER_UNION:
    try:
        return info.context.request.user  # type: ignore
    except AttributeError:
        return info.context["request"].user


def cast_to_status_user(user: USER_UNION) -> UserProto:
    user.status  # type: ignore  # raise attribute error
    return user  # type: ignore


def get_user_by_email(email: str) -> "UserProto":
    user = USER_MODEL.objects.get(**{USER_MODEL.EMAIL_FIELD: email})
    assert hasattr(user, "status")
    return user  # type: ignore


def get_token(user, action, **kwargs):
    username = user.get_username()
    if hasattr(username, "pk"):
        username = username.pk
    payload = {user.USERNAME_FIELD: username, "action": action}
    if kwargs:
        payload.update(**kwargs)
    token = signing.dumps(payload)
    return token


def get_payload_from_token(token, action, exp=None):
    payload = signing.loads(token, max_age=exp)
    _action = payload.pop("action")
    if _action != action:
        raise TokenScopeError
    return payload


def revoke_user_refresh_token(user):
    refresh_tokens = user.refresh_tokens.all()
    for refresh_token in refresh_tokens:
        with contextlib.suppress(Exception):
            refresh_token.revoke()


def fields_names(strawberry_fields: Iterable[StrawberryField]):
    return [field.python_name for field in strawberry_fields]


def inject_fields(fields: typing.Iterable[StrawberryField], annotations_only=False):
    def wrapped(cls: type):
        # python 3.8 compat:
        if not hasattr(cls, "__annotations__"):  # pragma: no cover
            cls.__annotations__ = {}

        for field in fields:
            if not field.name:
                continue
            if not annotations_only:
                setattr(cls, field.name, field)
            assert field.type_annotation
            cls.__annotations__[field.name] = field.type_annotation.annotation
        return cls

    return wrapped


def inject_arguments(args: Dict[str, type]):
    """Injects arguments to the decorated resolver.

    :param args:`dict[name, type]` of arguments to be injected.,
    """

    def wrapped(fn):
        sig = inspect.signature(fn)
        params = {
            inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=type_)
            for name, type_ in args.items()
        }
        params.update(sig.parameters.values())
        fn.__signature__ = inspect.signature(fn).replace(parameters=params)
        return fn

    return wrapped


def utc_now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)



from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import (
    PasswordChangeForm,
    UserChangeForm,
    UserCreationForm,
    UsernameField,
)

from gqlauth.core.utils import fields_names
from gqlauth.settings import gqlauth_settings as app_settings

REGISTER_MUTATION_FIELDS = fields_names(app_settings.REGISTER_MUTATION_FIELDS)


class RegisterForm(UserCreationForm):
    email = forms.EmailField(required="email" in REGISTER_MUTATION_FIELDS)

    class Meta:
        model = get_user_model()
        fields = REGISTER_MUTATION_FIELDS


class PasswordChangeFormGql(PasswordChangeForm):
    field_order = ["oldPassword", "newPassword1", "newPassword2"]


class EmailForm(forms.Form):
    email = forms.EmailField(max_length=254)


class CustomUsernameField(UsernameField):
    required = False


class UpdateAccountForm(UserChangeForm):
    class Meta:
        model = get_user_model()
        fields = fields_names(app_settings.UPDATE_MUTATION_FIELDS)
        field_classes = {"username": CustomUsernameField}


class PasswordLessRegisterForm(UserCreationForm):
    """A RegisterForm with optional password inputs."""

    class Meta:
        model = get_user_model()
        fields = REGISTER_MUTATION_FIELDS

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["password1"].required = False
        self.fields["password2"].required = False

    def save(self, commit=True):
        user = super().save(commit=False)
        user.set_unusable_password()
        if commit:
            user.save()
        return user



import warnings

from django.conf import settings as django_settings

from gqlauth.settings_type import GqlAuthSettings

gqlauth_settings: GqlAuthSettings

if user_settings := getattr(django_settings, "GQL_AUTH", False):
    if isinstance(user_settings, GqlAuthSettings):
        gqlauth_settings = user_settings

    else:
        raise Exception(
            f"GQL_AUTH settings should be of type "
            f"{GqlAuthSettings}, but you provided {type(user_settings)}"
        )

else:  # pragma: no cover
    warnings.warn("You have not provided any custom gql auth settings falling back to defaults")
    gqlauth_settings = GqlAuthSettings()



# Generated by Django 4.1.9 on 2023-05-07 16:06

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("gqlauth", "0003_delete_captcha"),
    ]

    operations = [
        migrations.CreateModel(
            name="Captcha",
            fields=[
                (
                    "uuid",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("text", models.CharField(editable=False, max_length=50)),
                ("insert_time", models.DateTimeField(auto_now_add=True)),
                ("tries", models.IntegerField(default=0)),
                (
                    "image",
                    models.ImageField(
                        editable=False,
                        help_text="url for the captcha image",
                        upload_to="captcha/%Y/%m/%d/",
                    ),
                ),
            ],
        ),
    ]



from enum import Enum
from typing import Optional, TypeVar

import strawberry
from graphql import GraphQLError

from gqlauth.core.scalars import ExpectedErrorType

T = TypeVar("T")


@strawberry.type
class MutationNormalOutput:
    success: bool
    errors: Optional[ExpectedErrorType] = None


class GQLAuthError(GraphQLError):
    def __init__(self, code: "GQLAuthErrors", *args, **kwargs):
        kwargs["message"] = ""
        super().__init__(*args, **kwargs)
        if not self.message:
            self.message = code.value


class GQLAuthErrors(Enum):
    UNAUTHENTICATED = "Unauthenticated."
    INVALID_TOKEN = "Invalid token."
    EXPIRED_TOKEN = "Expired token."
    NO_SUFFICIENT_PERMISSIONS = "Permissions found could not satisfy the required permissions."
    NOT_VERIFIED = "Please verify your account."
    MISSING_TOKEN = "No JWT found"



from uuid import UUID

import strawberry
import strawberry_django

from gqlauth.captcha import models
from gqlauth.captcha.models import PILLOW_INSTALLED
from gqlauth.core.scalars import image as Image

if PILLOW_INSTALLED:

    @strawberry_django.type(model=models.Captcha)
    class CaptchaType:
        uuid: UUID
        image: strawberry.auto

        @strawberry_django.field(description="returns the b64 encoded image.")
        def pil_image(self: models.Captcha) -> Image:  # type: ignore
            return self.as_bytes()  # type: ignore



import contextlib
import typing
from base64 import b64decode, b64encode

import strawberry

from gqlauth.core.exceptions import WrongUsage
from gqlauth.core.utils import camelize


def serialize_excpected_error(errors):
    if isinstance(errors, dict):
        if errors.get("__all__", False):
            errors["non_field_errors"] = errors.pop("__all__")
        return camelize(errors)
    elif isinstance(errors, list):
        return {"nonFieldErrors": errors}
    raise WrongUsage("`errors` must be list or dict!")


ExpectedErrorType = strawberry.scalar(
    typing.NewType("ExpectedError", dict),
    description="""
     Errors messages and codes mapped to
    fields or non fields errors.
    Example:
    {
        field_name: [
            {
                "message": "error message",
                "code": "error_code"
            }
        ],
        other_field: [
            {
                "message": "error message",
                "code": "error_code"
            }
        ],
        nonFieldErrors: [
            {
                "message": "error message",
                "code": "error_code"
            }
        ]
    }
    """,
    serialize=lambda value: serialize_excpected_error(value),
    parse_value=lambda value: value,
)

with contextlib.suppress(ImportError):
    from PIL.Image import Image

    image = strawberry.scalar(
        typing.NewType("image", Image),
        serialize=lambda v: b64encode(v).decode("ascii"),
        parse_value=lambda v: b64decode(v),
    )



from __future__ import annotations

import importlib.util
import io
import uuid

from django.db import models
from django.utils import timezone

from gqlauth.captcha.captcha_factorty import CaptchaInstanceType, generate_captcha_text
from gqlauth.core.constants import Messages
from gqlauth.settings import gqlauth_settings as app_settings

PILLOW_INSTALLED = False
if importlib.util.find_spec("PIL"):
    PILLOW_INSTALLED = True


class Captcha(models.Model):
    instance: CaptchaInstanceType
    uuid = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    text = models.CharField(max_length=50, editable=False)
    insert_time = models.DateTimeField(auto_now_add=True, editable=False)
    tries = models.IntegerField(default=0)

    objects: models.Manager[Captcha]

    if PILLOW_INSTALLED:
        image = models.ImageField(
            blank=False,
            null=False,
            upload_to="captcha/%Y/%m/%d/",
            editable=False,
            help_text="url for the captcha image",
        )

    @staticmethod
    def _format(text: str):
        return text.lower().replace(" ", "")

    @classmethod
    def create_captcha(cls):
        cap = generate_captcha_text()
        obj = cls(text=cap.text)
        # saving the image for future use when resolving to base64 or saving to .png
        obj.instance = cap
        obj.save()
        if app_settings.FORCE_SHOW_CAPTCHA:
            cap.pil_image.show()
        if app_settings.CAPTCHA_SAVE_IMAGE:
            django_content_file = obj.instance.to_django(name=str(obj.uuid))
            obj.image.save(django_content_file.name, django_content_file)

        return obj

    def save(self, *args, **kwargs):
        self.text = self._format(self.text)
        super().save(*args, **kwargs)

    def validate(self, user_entry: str):
        """Validates input_.

        - if tried to validate more than 3 times obj will be deleted in the database
        - else increments by one
        - if reaches expiery date deletes the obj
        - returns bool for success state
        """
        if self.tries > app_settings.CAPTCHA_MAX_RETRIES:
            try:
                self.delete()
            except ValueError:
                ...  # object is deleted
            return Messages.CAPTCHA_MAX_RETRIES

        self.tries += 1

        # check expiery
        if timezone.now() > self.insert_time + app_settings.CAPTCHA_EXPIRATION_DELTA:
            try:
                self.delete()
            except ValueError:
                ...  # object is deleted
            return Messages.CAPTCHA_EXPIRED

        # validate
        if app_settings.CAPTCHA_TEXT_VALIDATOR(
            self.text.replace(" ", ""), user_entry.replace(" ", "")
        ):
            # delete captcha if valid
            self.delete()
            return Messages.CAPTCHA_VALID

        return Messages.CAPTCHA_INVALID

    def as_bytes(self):
        """Stores the image on a bytes_array.

        The scalar will further convert it to b64 string representation.
        """
        bytes_array = io.BytesIO()
        self.instance.pil_image.save(bytes_array, format="PNG")
        return bytes_array.getvalue()

    def __str__(self):
        interval = (self.insert_time + app_settings.CAPTCHA_EXPIRATION_DELTA) - timezone.now()
        interval = interval.total_seconds()
        expiery_str = f" expires in {interval} seconds" if interval > 0 else "already expierd"
        return "captcha " + expiery_str



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Callable

from asgiref.sync import sync_to_async
from django.contrib.auth.models import AnonymousUser
from django.http import HttpRequest
from django.utils.decorators import sync_and_async_middleware
from strawberry import Schema

from gqlauth.core.exceptions import TokenExpired
from gqlauth.core.types_ import GQLAuthError, GQLAuthErrors
from gqlauth.core.utils import USER_UNION, app_settings
from gqlauth.jwt.types_ import TokenType
from jwt import PyJWTError

anon_user = AnonymousUser()
if TYPE_CHECKING:
    pass


class UserOrError:
    __slots__ = ("user", "error")

    def __init__(self, user: USER_UNION = anon_user, error: Exception | None = None):
        self.user = user
        self.error = error


USER_OR_ERROR_KEY = UserOrError.__name__


def get_user_or_error(scope_or_request: dict | HttpRequest) -> UserOrError:
    user_or_error = UserOrError()
    if token_str := app_settings.JWT_TOKEN_FINDER(scope_or_request):
        try:
            token = TokenType.from_token(token=token_str)
            user = token.get_user_instance()
            user_or_error.user = user

        except PyJWTError:  # raised by python-jwt
            user_or_error.error = GQLAuthError(code=GQLAuthErrors.INVALID_TOKEN)

        except TokenExpired:
            user_or_error.error = GQLAuthError(code=GQLAuthErrors.EXPIRED_TOKEN)

    else:
        user_or_error.error = GQLAuthError(code=GQLAuthErrors.MISSING_TOKEN)

    return user_or_error


def channels_jwt_middleware(inner: Callable):
    from channels.db import database_sync_to_async

    if asyncio.iscoroutinefunction(inner):
        get_user_or_error_async = database_sync_to_async(get_user_or_error)

        async def middleware(scope, receive, send):
            if not scope.get(USER_OR_ERROR_KEY, None):
                user_or_error: UserOrError = await get_user_or_error_async(scope)
                scope[USER_OR_ERROR_KEY] = user_or_error
            return await inner(scope, receive, send)

    else:  # pragma: no cover
        raise NotImplementedError("sync channels middleware is not supported yet.")
    return middleware


@sync_and_async_middleware
def django_jwt_middleware(get_response):
    def logic(request: HttpRequest) -> None:
        if not hasattr(request, USER_OR_ERROR_KEY):
            user_or_error: UserOrError = get_user_or_error(request)
            setattr(request, USER_OR_ERROR_KEY, user_or_error)

    if asyncio.iscoroutinefunction(get_response):
        async_logic = sync_to_async(logic)

        async def middleware(request: HttpRequest):
            await async_logic(request)
            return await get_response(request)

    else:

        def middleware(request: HttpRequest):  # type: ignore
            logic(request)
            return get_response(request)

    return middleware


class JwtSchema(Schema):
    """Injects token to context."""

    def execute_sync(self, *args, **kwargs):
        self._inject_user_and_errors(kwargs)
        return super().execute_sync(*args, **kwargs)

    async def execute(self, *args, **kwargs):
        self._inject_user_and_errors(kwargs)
        return await super().execute(*args, **kwargs)

    async def subscribe(self, *args, **kwargs):
        self._inject_user_and_errors(kwargs)
        return await super().subscribe(*args, **kwargs)

    @staticmethod
    def _inject_user_and_errors(kwargs: dict) -> UserOrError:
        context = kwargs.get("context_value")
        # channels compat
        if isinstance(context, dict):
            request = context["request"]
            try:
                user_or_error: UserOrError = request.consumer.scope[USER_OR_ERROR_KEY]
            except AttributeError:
                # FIXME: This is here because the test consumer is not 1:1 as a real client ATM
                user_or_error = request.scope[USER_OR_ERROR_KEY]
            request.user = user_or_error.user  # type: ignore
        else:
            user_or_error: UserOrError = getattr(context.request, USER_OR_ERROR_KEY)  # type: ignore
            context.request.user = user_or_error.user  # type: ignore
        return user_or_error



# Generated by Django 4.1.7 on 2023-05-07 15:38

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("gqlauth", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="userstatus",
            options={"verbose_name_plural": "User statuses"},
        ),
    ]



# Generated by Django 4.1.9 on 2023-05-07 15:59

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("gqlauth", "0002_alter_userstatus_options"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Captcha",
        ),
    ]
